"""OLPC Contents manifest generation utilities."""
# Copyright (C) 2007 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
from __future__ import with_statement

import hashlib, os, os.path, stat, unicodedata
import bitfrost.util.json as json
try:
    from xattr import listxattr, getxattr, setxattr
except ImportError:
    def listxattr(filename,symlink=False):
        """Local alias for `xattr.listxattr`."""
        return ()

DIROBJECT = '.xo-files'
"""Name of the per-directory contents file when using `SplitContents`."""
HASHLIST = [ 'sha-256', 'ripemd-160' ]
"""Ordered list of the hashes used for contents files."""

def mkhashes(content=None, filename=None):
    """Make hashes for a directory object."""
    h1 = hashlib.sha256()
    h2 = hashlib.new('ripemd160')
    if filename is None:
        h1.update(content)
        h2.update(content)
    else:
        BLOCK_SIZE=65536
        with open(filename, 'r', 0) as f:
            while True:
                buf = f.read(BLOCK_SIZE)
                if not buf: break
                h1.update(buf)
                h2.update(buf)
    return [ h1.hexdigest(), h2.hexdigest() ]

def _checkUTF8(s):
    """Check that the given string is valid UTF-8 in Normalization Form C."""
    try:
        if type(s) is unicode:
            # Unicode is not UTF-8, unless it happens to contain only ASCII.
            return s.encode('ASCII')
        u = s.decode('utf-8')
        ns = unicodedata.normalize("NFC", u).encode('utf-8')
        if s != ns:
            raise RuntimeError('File, directory, symlink, user, or group name '
                               'not NFC UTF-8: %s (should be %s)' \
                               % (repr(s), repr(ns)))
        return s # okay!
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise RuntimeError('File, directory, symlink, user, or group name not '
                           'NFC UTF-8: %s' % repr(s))

def mkusermap(passwd_file='/etc/passwd'):
    """Make a map from UID to username from a passwd-format file."""
    usermap = {}
    with open(passwd_file) as f:
        for line in f:
            username, passwd, uid, gid, comment, homedir, shell=line.split(':')
            usermap[int(uid)] = username
    return usermap

def mkgroupmap(group_file='/etc/group'):
    """Make a map from GID to groupname from a file in the format of
    /etc/group."""
    groupmap = {}
    with open(group_file) as f:
        for line in f:
            groupname, passwd, gid, userlist = line.split(':')
            groupmap[int(gid)] = groupname
    return groupmap

def mkdirobject (path, accum=None, write=False,
                 usermap=None, groupmap=None,
                 force_user=None, force_group=None,
                 valid_UTF8=True):
    """Makes a directory object for the directory given by `path`.

    Returns a pair (enveloped object, contents length) where the
    'contents length' element is the integer length of a canonical
    JSON-encoded contents manifest for this directory and its children, if one
    were to be created.  Contents length = 16 + sum(length of d_i +
    1), where the d_i are the directory objects for this directory and
    all of its children."""
    check = _checkUTF8 if valid_UTF8 else lambda s: s
    if not os.path.isdir(path): raise RuntimeError("not a directory: %s"%path)
    if accum is None: accum = []

    contents_map = {}
    length = 16

    for f in reversed(sorted(os.listdir(path))):
        if f == DIROBJECT: continue
        fullname = os.path.join(path, f)
        si = os.lstat(fullname)
        entry = { 'm': si.st_mode, 'u#': si.st_uid, 'g#': si.st_gid }
        # force user and group ids.
        if type(force_user) is int:
            entry['u#'] = force_user
        if type(force_group) is int:
            entry['g#'] = force_group
        # look up 'u' attribute (textual username)
        if usermap and entry['u#'] in usermap:
            entry['u'] = check(usermap[entry['u#']])
        # look up 'g' attribute (textual groupname)
        if groupmap and entry['g#'] in groupmap:
            entry['g'] = check(groupmap[entry['g#']])
        # force user and group names if they weren't integral
        if force_user is not None and type(force_user) is not int:
            entry['u'] = check(force_user)
            del entry['u#']
        if force_group is not None and type(force_group) is not int:
            entry['g'] = check(force_group)
            del entry['g#']
        # okay, deal with the different file types.
        if stat.S_ISDIR(si.st_mode):
            subdir_name = os.path.join(fullname, DIROBJECT)
            json_subdir, subdir, m_length = \
                         mkdirobject(fullname, accum, write,
                                     usermap, groupmap,
                                     force_user, force_group, valid_UTF8)
            length += m_length - 16
            entry['h'] = mkhashes(json_subdir)
            entry['dl'] = len(json_subdir)
            entry['ml'] = m_length
        elif stat.S_ISREG(si.st_mode):
            entry['h'] = mkhashes(filename=fullname)
        elif stat.S_ISCHR(si.st_mode) or stat.S_ISBLK(si.st_mode):
            entry['d'] = si.st_rdev
        elif stat.S_ISLNK(si.st_mode):
            entry['l'] = check(os.readlink(fullname))
        elif stat.S_ISFIFO(si.st_mode) or stat.S_ISSOCK(si.st_mode):
            pass # no special info
        else:
            raise RuntimeError(path) # odd file!

        # xattr support: disabled, because we have shipped a fv for many
        # releases that views 'x' as invalid data. this could be re-enabled
        # with a future incompatible change to the contents format.
        """
        xattr = sorted([n for n in listxattr(fullname, True)])
        xattr = [(n, getxattr(fullname, n, True)) for n in xattr]
        if len(xattr) > 0:
            entry['x'] = xattr
        """

        # note that filenames should be in UTF-8
        contents_map[check(f)] = entry

    directory_object = make_envel('dir', 1, [ HASHLIST, contents_map ])
    accum.append(directory_object) # directories in reversed post-order
    jsoned_directory_object = json.write(directory_object)
    length += len(jsoned_directory_object) + 1
    if write:
        # write directory object file in Canonical JSON format
        with open(os.path.join(path, DIROBJECT), 'w') as f:
            f.write(jsoned_directory_object)

    return jsoned_directory_object, directory_object, length

def mkcontents (path, write_dir=False, usermap=None, groupmap=None,
                force_user=None, force_group=None, valid_UTF8=True):
    """Make a contents manifest for the directory given by `path`.

    Returns the enveloped object."""

    if not os.path.isdir(path): raise RuntimeError

    directory_list = []
    mkdirobject(path, directory_list, write=write_dir,
                usermap=usermap, groupmap=groupmap,
                force_user=force_user, force_group=force_group,
                valid_UTF8=valid_UTF8)
    directory_list.reverse()
    # trim contents if we're generating a split one
    if write_dir:
        directory_list = directory_list[0:1]
        
    # emit complete contents manifest in Canonical JSON format.
    return make_envel('manifest', 1, directory_list)

##### abstract away whether contents manifest is scattered in .xo-files or
##### in a single unified contents file.
class Contents(object):
    """Abstract superclass of `UnifiedContents` and `SplitContents`."""
    def contents_hash(self):
        """Return a string which is a hash over the entire contents."""
        return mkhashes(self[''])[0]
    def get_dirobject_pair(self, path):
        """Returns a dirobject as a pair.  The first component is the
        json-encoded dirobject, and the second component is the native
        python object.  Use this method when you need both
        representations of the dirobject; use `__getitem__` if you
        only need the json-encoded dirobject."""
        json_dir = self[path]
        return (json_dir, json.read(json_dir))
    def __getitem__(self, relative_dir_path):
        """Return the canonical JSON-encoded directory object
        corresponding to the (relative) directory named."""
        raise NotImplementedError # subclasses must implement

class UnifiedContents(Contents):
    """Read directory objects from a single unified contents manifest.  Seeks
    through the file to avoid having to load the entire contents manifest into
    memory.  Parent directory objects must be read before their children."""
    def __init__(self, contents_path):
        """Create a UnifiedContents object given the pathname to a
        unified contents manifest file."""
        self.file = open(contents_path, 'r')
        # offset for root.
        self.offset = { '': 15 } # opening '[' of root directory object
        # root is tricky: how big is it?  need to read incrementally. =(
        if self.file.read(15) != '["manifest",1,[':
            raise VerificationError('Bad contents manifest')
        # match [] to find end of root directory object.
        brackets = 0
        instring = False
        while True:
            c = self.file.read(1)
            if c=='': raise VerificationError('Short contents manifest')
            elif c=='\\' and instring:
                self.file.read(1) # throw away escaped char
            elif c=='"':
                instring = not instring # enter/leave quoted string
            elif c=='[' and not instring: brackets += 1
            elif c==']' and not instring: brackets -= 1
            if brackets == 0: break
        self.dlength = { '': self.file.tell()-15 }
        self.parsed = set()
    def _getitem(self, relative_dir_path, return_pair=False):
        """Implementation method for `__getitem__` and `get_dirobject_pair`,
        depending on the value of the `return_pair` argument."""
        self.file.seek(self.offset[relative_dir_path])
        json_dir = self.file.read(self.dlength[relative_dir_path])
        if relative_dir_path in self.parsed:
            return (json_dir, json.read(json_dir)) if return_pair else json_dir
        # now compute offsets and lengths for subdirectories.
        offset = self.file.tell() + 1 # skip trailing comma
        dirobject = json.read(json_dir)
        hashes, entries = open_envel('dir',1,dirobject)
        for name in sorted(entries.keys()):
            props = entries[name]
            if stat.S_ISDIR(props['m']):
                path = os.path.join(relative_dir_path, name)
                self.dlength[path] = props['dl']
                self.offset[path] = offset
                offset += props['ml'] - 16
        self.parsed.add(relative_dir_path)
        # groovy. return directory object.
        return (json_dir, dirobject) if return_pair else json_dir
    def __getitem__(self, relative_dir_path):
        return self._getitem(relative_dir_path, False)
    def get_dirobject_pair(self, relative_dir_path):
        return self._getitem(relative_dir_path, True)

class SplitContents(Contents):
    """Read split directory objects from .xo-files in a directory tree."""
    def __init__(self, split_root):
        """Create a SplitContents given a root from which to fetch the
        split directory object .xo-files files."""
        self.split_root = split_root
    def __getitem__(self, relative_dir_path):
        return open(os.path.join(self.split_root, relative_dir_path,
                                 DIROBJECT), 'r').read()

##### Contents Manifest validation/verification #####
class VerificationError(Exception):
    """Exception thrown for verification problems."""
    pass
class EnvelopeError(VerificationError):
    """Exception thrown for envelope problems."""
    pass

def verify (root, path, contents, check_dlen=None, check_mlen=None,
            allow_extra=False, check_hash=None, strict_owner=False,
            recursive=True, usermap=None, groupmap=None):
    """Verify the directory object in the directory given by `path`.
    Returns the hashes of the directory object."""
    fullpath = os.path.join(root, path)
    if not os.path.isdir(fullpath):
        raise VerificationError(fullpath+' is not a directory')
    json_dir, directory_object = contents.get_dirobject_pair(path)
    dir_hash = mkhashes(json_dir)
    dir_length = len(json_dir)
    if check_dlen is not None and check_dlen != dir_length:
        raise VerificationError("Directory object length mismatch: "+fullpath)
    if check_hash is not None and check_hash != dir_hash:
        raise VerificationError("Directory hash mismatch: "+fullpath)
    # check envelope format.
    hash_list, contents_map = open_envel('dir', 1, directory_object)
    # check hashes list
    if hash_list != HASHLIST:
        raise VerificationError("Bad hash list: " + hash_list)

    # check that directory contents match contents manifest.
    namelist = sorted(contents_map.keys())
    if namelist != sorted([f for f in os.listdir(fullpath) if f!=DIROBJECT]):
        if not allow_extra:
            raise VerificationError("Directory listings differ: "+fullpath)
    namelist = None # free memory
    dirlist = []
    mlen = 16 + dir_length + 1
    for name, entry in contents_map.items():
        fullname = os.path.join(fullpath, name)
        si = os.lstat(fullname)
        if si.st_mode != entry['m']:
            raise VerificationError("Mode mismatch: "+fullname)
        if strict_owner:
            if 'u' not in entry or 'u#' not in entry:
                raise VerificationError("Owner info missing from entry: " +
                                        fullname)
            if 'g' not in entry or 'g#' not in entry:
                raise VerificationError("Group info missing from entry: "+
                                        fullname)
        if 'u#' in entry and si.st_uid != entry['u#']:
            raise VerificationError("Owner # mismatch: "+fullname)
        if 'g#' in entry and si.st_gid != entry['g#']:
            raise VerificationError("Group # mismatch: "+fullname)
        if 'u' in entry:
            if usermap is None:
                raise VerificationError("Symbolic username: "+fullname)
            if usermap[si.st_uid] != entry['u']:
                raise VerificationError("Owner name mismatch: "+fullname)
        if 'g' in entry:
            if groupmap is None:
                raise VerificationError("Symbolic group name: "+fullname)
            if groupmap[si.st_gid] != entry['g']:
                raise VerificationError("Group name mismatch: "+fullname)
        # check xattrs
        xattr = entry.get('x', [])
        if ([x[0] for x in xattr] !=
            sorted([n for n in listxattr(fullname, True)])):
            raise VerificationError("xattr name mismatch: "+fullname)
        if ([x[1] for x in xattr] !=
            [getxattr(fullname, x[0], True) for x in xattr]):
            raise VerificationError("xattr value mismatch: "+fullname)
        
        mode = entry['m']
        good_props = [ 'm', 'u', 'u#', 'g', 'g#', 'x' ]
        if stat.S_ISDIR(mode):
            good_props += [ 'h', 'dl', 'ml' ]
            dirlist.append((os.path.join(path, name), entry))
            mlen += (entry['ml']-16)
        elif stat.S_ISREG(mode):
            good_props += [ 'h' ]
            if mkhashes(filename=fullname) != entry['h']:
                raise VerificationError("Content hash verification error: " +
                                        fullname)
        elif stat.S_ISCHR(mode) or stat.S_ISBLK(mode):
            good_props += [ 'd' ]
            if si.st_rdev != entry['d']:
                raise VerificationError("Device node mismatch: "+fullname)
        elif stat.S_ISLNK(mode):
            good_props += [ 'l' ]
            if os.readlink(fullname) != entry['l']:
                raise VerificationError("Symlink mismatch: "+fullname)
        elif stat.S_ISFIFO(mode) or stat.S_ISSOCK(mode):
            pass # no additional properties to check
        else:
            raise VerificationError("Unknown file mode: "+fullname)
        for k in entry.keys():
            if k not in good_props:
                raise VerificationError('Bad property %s in %s' %
                                        (k, fullname))

    # check contents manifest length
    if check_mlen is not None and check_mlen != mlen:
        raise VerificationError("Contents manifest length mismatch: "+fullpath)
    # ok, recurse: check that directory contents manifests verify
    if recursive:
        for d, entry in dirlist:
            verify(root, d, contents, entry['dl'], entry['ml'],
                   allow_extra, entry['h'], strict_owner, recursive,
                   usermap, groupmap)
    return dir_hash
    
##### clean files not mentioned in contents manifest
def _forceremove (filename):
    """/bin/rm -rf filename"""
    st = os.lstat(filename)
    if stat.S_ISDIR(st.st_mode):
        for f in os.listdir(filename):
            _forceremove(os.path.join(filename, f))
        os.rmdir(filename)
    else:
        os.remove(filename)

def clean (root, path, contents, dryrun=False, valid_UTF8=True):
    """Remove files not mentioned in the contents manifest in `path`."""

    check = _checkUTF8 if valid_UTF8 else lambda s: s
    fullpath = os.path.join(root, path)
    __, dirobject = contents.get_dirobject_pair(path)
    __, contents_map = open_envel('dir', 1, dirobject)
    files = {}
    for name, entry in contents_map.items():
        files[name] = entry['m']
    for f in os.listdir(fullpath):
        check(f) # ensure we don't delete a file just because encoding is wrong
        if f == DIROBJECT: continue
        if f in files:
            if stat.S_ISDIR(files[f]):
                clean(root, os.path.join(path, f), contents, dryrun, valid_UTF8)
        else:
            if dryrun: print os.path.join(fullpath, f)
            else: _forceremove(os.path.join(fullpath, f))

##### simple envelope parsing / generation
def open_envel(typename, version, envelope):
    """Simple helper to unpack an enveloped value."""
    e_ty, e_version, e_data = envelope
    if e_ty != typename:
        raise EnvelopeError("Envelope has wrong type: %s" % e_ty)
    if e_version != version:
        raise EnvelopeError("Envelope has wrong version: %d" % e_version)
    return e_data

def make_envel(typename, version, envelope):
    """Trivial helper to create an enveloped value."""
    return (str(typename), int(version), envelope)

########## BACKWARDS COMPATIBILITY #############
mkmanifest = mkcontents
"""Alias for `mkcontents` for backwards-compatibility.

:deprecated: Use `mkcontents` instead."""
Manifest = Contents
"""Alias for `Contents` for backwards-compatibility.

:deprecated: Use `Contents` instead."""
UnifiedManifest = UnifiedContents
"""Alias for `UnifiedContents` for backwards-compatibility.

:deprecated: Use `UnifiedContents` instead."""
SplitManifest = SplitContents
"""Alias for `SplitContents` for backwards-compatibility.

:deprecated: Use `SplitContents` instead."""
