#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2007-2008 One Laptop Per Child Association, Inc.
# Written by Ivan Krstić and C. Scott Ananian

# This program is free software; you can redistribute it and/or modify
# it under the terms of version 2 (and no other version) of the GNU
# General Public License as published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""Bitfrost lease framework.

This file defines the basic parameters for bitfrost anti-theft leases.

The lease data is a Canonical JSON array composed of an integer and a
dictionary, in that order. The integer is the version identifier for
the lease format, and the dictionary consists of one or more leases,
where a machine's serial number (which is in actuality a string) is
the key, and the value is interpreted according to the version
identifier.

Version 1 of the lease format is documented at
http://wiki.laptop.org/go/Firmware_Key_and_Signature_Formats#Antitheft.2FActivation_Lease
"""

import bitfrost.util.json as json
from bitfrost import _assert
from bitfrost.leases.crypto import verify_act
from bitfrost.leases.errors import *
from bitfrost.leases.keys import LEASE_KEYS
import re

def find_lease (this_sn, this_uuid, lease_contents):
    """Attempt to interpret the lease data and find this computer's lease,
    then cryptographically verify it.  If it validates, return the
    'minimal' lease, otherwise raise an exception."""
    
    # 'version' here is the version of the lease file encapsulation, that is,
    # the version of the serial-number-to-lease data structure.  It should
    # only be changed if the means we use to marshall multiple leases into a
    # file changes.  Thus, 'version' is *not* passed into verify_act; the
    # lease versioning is done separately inside verify_act.
    try:
        version, leases = json.read(lease_contents)
    except:
        # perhaps this is a singleton lease
        version, leases = 1, { this_sn: lease_contents }
    try:
        _assert(isinstance(version, int))
        if version != 1:
            # version of the multiple-lease marshalling, *not* the version
            # of the lease itself.
            raise UnsupportedLeaseFormat(version)
        _assert(isinstance(leases, dict))
        _assert(len(leases.keys()) > 0)
    except:
        raise InvalidLeaseData()

    if not this_sn in leases:
        raise NoLeaseFound()

    for l in re.split('(?:\r?\n)|\r', leases[this_sn]):
        try:
            if verify_act(this_sn, this_uuid, l+'\n',
                            LEASE_KEYS):
                return l+'\n' # minimized lease
        except: pass # try the next lease

    raise VerificationFailure()
