"""Use contents manifest to improve rsync performance by only syncing
directories which are out-of-date."""
# Copyright (C) 2007-8 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>

from __future__ import with_statement
import os, os.path, shutil, stat, sys
import time
import subprocess
from bitfrost.contents.utils import mkhashes, open_envel, UnifiedContents
from bitfrost import _assert
from subprocess import check_call
HASH, MODE, UID, GID, XATTR = 'h', 'm', 'u#', 'g#', 'x'

def rsync (src, dst, path, recursive, verbosity=0, link_dest=None, tries=3, is_dir=True):
    """Invoke rsync with appropriate options."""

    if path.startswith('/'):
        path = path.lstrip('/')
    src_uri = os.path.join(src, path)
    if is_dir:
        src_uri = src_uri + '/'

    # the '-c' option is a bit sketch; we'd like to use filetimes, but my
    # current archive mechanism (untar all files in a build) doesn't allow it.
    # Have to be careful that build/archive mechanism maintains timestamps if
    # we want to use them.
    args = ['/usr/bin/rsync', '-aOyz', '-c', '--delete-during', '--numeric-ids',
            src_uri, os.path.join(dst, path)]
    if src_uri.startswith("rsync://"):
        args[2:2] = ['--timeout=180', '--contimeout=180']
    if not recursive:
        args[2:2] = ['--no-r', '--dirs']
    if link_dest is not None:
        link_path = os.path.join(link_dest, path)
        # rsync will complain if we give it a link-dest which doesn't actually
        # exist, isn't a directory, or if the link-dest is relative.
        _assert(os.path.isabs(link_path))
        if os.path.isdir(link_path):
            args[2:2] = [ '--link-dest', link_path ]
    if verbosity > 1: args[2:2] = ['-v']
    if verbosity > 2: args[2:2] = ['--itemize-changes']
    if verbosity > 3: args[2:2] = ['-v']

    for i in range(0, tries):
        if verbosity > 0: print "Rsyncing: %s (try %d/%d)" % (path, i + 1, tries)
        try:
            check_call(args)
            break
        except subprocess.CalledProcessError:
            if i == tries - 1:
                raise
            else:
                if verbosity > 0: print "Sync failed, retrying in 10 seconds."
                time.sleep(10)

def sync_and_check (src, link, dst, path, old_contents, new_contents,
                    verbosity=0):
    """Do an rsync on the dirty directory given by 'path'.  Afterwards,
    check the .rmanifest to see if any of the other directories need
    syncing, and if so, do it."""

    rsync(src, dst, path, False, verbosity, link_dest=link)
    __, dirobject = new_contents.get_dirobject_pair(path)
    __, contents_map = open_envel('dir', 1, dirobject)
    for name, entry in contents_map.items():
        if stat.S_ISDIR(entry[MODE]):
            # in non-recursive mode, rsync creates the directory but doesn't
            # fill it.
            npath = os.path.join(path, name)
            try:
                dm = old_contents[npath]
                if entry[HASH] != mkhashes(dm):
                    raise RuntimeError("hash doesn't match")
                # hashes match, so cp -al old to new
                shutil.rmtree(os.path.join(dst, npath), ignore_errors=True)
                check_call(['/bin/cp', '-al',
                            os.path.join(link, npath),
                            os.path.join(dst, npath)])
                # the hash of the dir was the same, but the owner or
                # permissions might differ.  Fix up.
                os.lchown(os.path.join(dst, npath), entry[UID], entry[GID])
                os.chmod(os.path.join(dst, npath), entry[MODE])
                if XATTR in entry:
                    # cheat and just reinvoke rsync to transfer xattrs.
                    rsync(src, dst, path, False, verbosity, link_dest=link)
            except: # hash mismatch (or some other problem)
                sync_and_check(src, link, dst, npath,
                               old_contents, new_contents, verbosity)

## Updaters built on the above primitives.

def _get_contents(url, contents):
    """Helper function to fetch the contents file from the given base url."""
    check_call(['rsync','-z',url+'/contents', contents])

def _cleanup_isolation(src, dst, verbosity):
    """Helper function to workaround imperfect isolation of /pristine.
    See http://dev.laptop.org/ticket/5051"""
    # we also need to fixup /etc/localtime and /etc/avahi/etc/localtime =(
    # (trac #8190)
    if verbosity >= 0: print >>sys.stderr, " - Cleaning up (trac #5051)"
    for f in [ '/etc/group', '/etc/group-', '/etc/passwd', '/etc/passwd-',
               '/etc/gshadow', '/etc/gshadow-', '/etc/shadow', '/etc/shadow-',
               '/etc/localtime']:
        rsync(src, dst, f, False, verbosity, is_dir=False)
    # NASTY HACK: since we don't have COW to protect our pristine tree,
    #   /var/run/utmp and /var/log/wtmp are being stomped on.
    # We also don't protect /etc/alsa/asound.state =(
    # and etc/avahi seems to be writable? (trac #8190)
    for hack_dir in [ 'var', 'etc/alsa', 'etc/avahi' ]:
        rsync(src, dst, hack_dir, True, verbosity)
    

def olpc_update_irsync_pristine(vhash, old_root, old_contents,
                                new_root, new_contents,
                                tmpdir, url, verbosity, **extra):
    """Do an incremental rsync, starting from a pristine tree."""
    if verbosity >=0: print >>sys.stderr, " - Fetching contents."
    shutil.copy(old_contents, new_contents)
    _get_contents(url, new_contents)
    old_contents_obj = UnifiedContents(old_contents)
    new_contents_obj = UnifiedContents(new_contents)
    old_contents_obj[''] # side-effect: load subdirs
    if verbosity >=0: print >>sys.stderr, " - Performing incremental rsync."
    sync_and_check(url+'/root', old_root, new_root, '',
                   old_contents_obj, new_contents_obj,
                   verbosity)
    _cleanup_isolation(url+'/root', new_root, verbosity)

def olpc_update_irsync_dirty(vhash, old_root, old_contents,
                             new_root, new_contents,
                             tmpdir, url, verbosity, **extra):
    """Do an incremental rsync, not assuming that tree is pristine."""
    if verbosity >=0: print >>sys.stderr, " - Creating contents for existing tree."
    # throw away the old_contents we are given, and make our own.
    old_contents = os.path.join(tmpdir, 'contents')
    check_call(['/usr/sbin/olpc-contents-create','-f', old_contents,
                '-p','%s/etc/passwd'%old_root, '-g', '%s/etc/group'%old_root,
                old_root])
    # okay, now invoke the pristine irsync
    olpc_update_irsync_pristine(vhash, old_root, old_contents,
                                new_root, new_contents,
                                tmpdir, url, verbosity, **extra)

def olpc_update_rsync(vhash, old_root, old_contents, new_root, new_contents,
                      tmpdir, url, verbosity, **extra):
    """Do a full rsync."""
    if verbosity >=0: print >>sys.stderr, " - Fetching contents."
    shutil.copy(old_contents, new_contents)
    _get_contents(url, new_contents)
    if verbosity >=0: print >>sys.stderr, " - Performing full rsync."
    rsync(url+'/root', new_root, '', True, verbosity, link_dest=old_root)
    # attempt to minimize effects of race by rsyncing certain files again
    # at the end.
    _cleanup_isolation(url+'/root', new_root, verbosity)
