"""OLPC updater; setup and update utility functions."""
# Copyright (C) 2007-8 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
from __future__ import with_statement
from tempfile import mkdtemp
from shutil import rmtree, copytree
from subprocess import call, check_call, Popen, PIPE
import os, os.path, re, struct, sys, threading, time
from bitfrost.leases.crypto import verify_sig, verify_dev, check_expiration_func
from bitfrost import _assert, find_devicetree

def _is_partitioned():
    # see if / is a different device from /bootpart
    if not os.path.isdir('/bootpart'): return False
    return os.stat('/').st_dev != os.stat('/bootpart').st_dev

def _cleanup_lock(filename):
    """Try to delete lock file if the PID which wrote it is no longer around.
    Return `True` if we were successful cleaning up the lock (which means
    it's worthwhile to retry."""
    try:
        oldpid = int(open(filename, 'r').read().strip())
        try:
            os.getpgid(oldpid)
            return False # process still exists.
        except OSError:
            # pid is dead, clean up lock & retry.
            os.unlink(filename)
            time.sleep(1) # limit busy looping even in worst case.
            return True # retry.
    except:
        # some other error: probably file has something other than a PID
        return False
    
# directory exists if powerd is running.  create a file here,
# named after our pid, to inhibit suspend.
POWERD_INHIBIT_DIR = '/var/run/powerd-inhibit-suspend'

_suspend_inhibit_count=0
_suspend_inhibit_lock=threading.RLock()
def _inhibit_and_return_uninhibit():
    """Ask OHM (via dbus) to inhibit suspend, and if we're successful,
    return a thunk to reverse this operation.  If we're not successful
    inhibiting suspend, then return a thunk which does nothing.
    Alternately, tell powerd to inhibit suspend."""

    def powerd_running():
        return os.access(POWERD_INHIBIT_DIR, os.W_OK)

    def powerd_inhibit(path):
        fd = open(path, 'w')
        fd.close()

    def powerd_uninhibit(path):
        try:
            os.unlink(path)
        except OSError:
            pass

    # support local nesting.
    def update_count(val, before=(lambda: None), after=(lambda: None)):
        global _suspend_inhibit_count, _suspend_inhibit_lock
        with _suspend_inhibit_lock:
            # an exception in before() won't leave the count incremented
            if _suspend_inhibit_count == 0: before()
            _suspend_inhibit_count += val
            if _suspend_inhibit_count <= 0: after()


    if powerd_running():
        inhib_path = POWERD_INHIBIT_DIR + "/%u" % os.getpid()
        update_count(1, before=(lambda: powerd_inhibit(inhib_path)))
        return lambda: update_count(-1,
                after=(lambda: powerd_uninhibit(inhib_path)))

    # else, it's ohmd:
    try:
        import dbus
        bus = dbus.SystemBus()
        proxy = bus.get_object('org.freedesktop.ohm',
                               '/org/freedesktop/ohm/Keystore')
        keystore = dbus.Interface(proxy, 'org.freedesktop.ohm.Keystore')
        update_count(1, before=(lambda: keystore.SetKey('suspend.inhibit', 1)))
        # we've successfully inhibited suspend.
        return lambda: update_count(-1,
                after=(lambda: keystore.SetKey('suspend.inhibit', 0)))
    except dbus.DBusException, e:
        # complain.
        print >>sys.stderr, "Error setting OHM inhibit: %s" % e
        # we didn't inhibit, so no need to unset later.
        return lambda : None

def inhibit_suspend(f):
    """Ask ohmd or powerd to prevent suspend during an update."""
    # see http://wiki.laptop.org/go/Ohm and dlo trac #7910
    def new_f(*args, **kwds):
        undo = _inhibit_and_return_uninhibit()
        try:
            return f(*args, **kwds)
        finally:
            undo()
    return new_f

def _mklockfile(filename):
    """Attempt to create a lock file; throws an exception if the lock file
    already exists and the process which created the lock file is still
    alive."""
    # This is broken for NFS; see 'man 2 open' for recommended procedure
    # if that is ever an issue.
    try:
        fd = os.open(filename, os.O_CREAT|os.O_EXCL|os.O_WRONLY)
        os.write(fd, str(os.getpid()))
        os.close(fd)
        return True
    except:
        # if file already exists, but pid doesn't exist anymore,
        # unlink and try again.
        if _cleanup_lock(filename):
            return _mklockfile(filename)
        # no, old lock file is legit.
        raise RuntimeError('Process %s already holds lockfile %s' %
                           (open(filename).read(), filename))

def with_lock_file(filename):
    """Decorator to limit concurrency of a given function using a lock file."""
    def apply_lock(f):
        def new_f(*args, **kwds):
            _mklockfile(filename)
            try:
                return f(*args, **kwds)
            finally:
                os.unlink(filename)
            new_f.func_name = f.func_name
        return new_f
    return apply_lock

def shorten_hash(vhash):
    """Return user-friendly short form of the version hash."""
    return vhash[0:32]

def rewrite_symlink(src, dst):
    """Atomically swing a symlink at dst from its current contents to src."""
    tmpdir = mkdtemp(prefix='sym', dir=os.path.dirname(dst))
    tmplnk = os.path.join(tmpdir, 'symlink')
    try:
        os.symlink(src, tmplnk)
        os.rename(tmplnk, dst)
    finally:
        rmtree(tmpdir, ignore_errors=True)

def trim_config(partitioned, cur_hash):
    """Remove 'alt' from current boot config."""
    if not partitioned:
        swing_config(partitioned, current=cur_hash, alt=None)
    else:
        #0-2. Make /boot/alt point to ../$cur_hash
        rewrite_symlink("../%s" % cur_hash, "/bootpart/boot/alt")

def swing_config(partitioned, current, alt=None):
    """Create new boot configuration."""
    if not partitioned:
        # 11. Make a new config /versions/configs/$d (d = mkdtemp)
        cfgdir = mkdtemp(prefix='cfg.', dir='/versions/configs/')
        # 12. Create 'current' symlink to ../../pristine/<hash>
        # XXX: should be ../../run/%s (trac #4898)
        os.symlink('/versions/pristine/%s' % current,
                   os.path.join(cfgdir, 'current'))
        # 13. Create 'alt' symlink to *realpath of* /versions/running
        if alt is not None:
            # XXX: should be ../../run/%s (trac #4898)
            os.symlink('/versions/pristine/%s' % alt,
                       os.path.join(cfgdir, 'alt'))
        # 14. Swing /versions/boot to /versions/configs/$d
        #      (atomic!  iff we do file move of new symlink)
        oldcfg = os.path.basename(os.readlink('/versions/boot').rstrip('/'))
        rewrite_symlink('configs/%s' % os.path.basename(cfgdir.rstrip('/')),
                        '/versions/boot')
        # 15. Delete /versions/configs/$c
        rmtree('/versions/configs/%s' % oldcfg)
    else:
        # 11. Copy /versions/pristine/<hash>/boot to boot:/boot-versions/<hash>
        bootver = "/bootpart/boot-versions/%s" % current
        rmtree(bootver, ignore_errors=True)
        copytree(src="/versions/pristine/%s/boot" % current,
                 dst="/bootpart/boot-versions/%s" % current, symlinks=True)
        # 12. Make boot:/boot-versions/<hash>/alt point to what boot:/boot currently points to
        if alt is not None:
            os.symlink('../%s' % alt, '/bootpart/boot-versions/%s/alt' % current)
        # 13. Atomically swing boot:/boot to point to /boot-versions/<hash>
        rewrite_symlink('boot-versions/%s' % current, '/bootpart/boot')

def clean_versions(report, partitioned, cur_hash, new_hash, clean_run=True, clean_configs=True):
    """Free up space for update by removing old or incomplete versions."""
    def rmrf(f, ignore_errors=False):
        if os.path.isdir(f) and not os.path.islink(f):
            if ignore_errors:
                call(['/bin/rm','-rf',f])
            else:
                check_call(['/bin/rm','-rf',f])
        else:
            try:
                os.unlink(f)
            except:
                if not ignore_errors: raise
    # 3a. Clean up pristine and running trees.
    # XXX: can we add a sanity check here by looking at '/'?
    keep = [ cur_hash, shorten_hash(cur_hash) ]
    try:
        keep.extend(os.listdir('/versions/sticky'))
    except: pass # if /versions/sticky doesn't exist, don't worry about it.

    for f in os.listdir('/versions/pristine'):
        _assert(len(f) > 0 and '.' not in f) # defensive programming.
        if f in keep: continue
        report(0, 'Deleting old pristine version %s' % f)
        rmrf('/versions/pristine/%s' % f, ignore_errors=True)

    if clean_run:
        for f in os.listdir('/versions/run'):
            _assert(len(f) > 0 and '.' not in f) # defensive programming.
            if f in keep: continue
            report(0, 'Deleting old run version %s' % f)
            rmrf('/versions/run/%s' % f, ignore_errors=True)

    for f in os.listdir('/versions/contents'):
        _assert(len(f) > 0 and '.' not in f) # defensive programming.
        if f in keep: continue
        report(0, 'Deleting old contents for version %s' % f)
        rmrf('/versions/contents/%s' % f, ignore_errors=True)
    # 3b. Delete corresponding members of /boot-versions if using a boot partition
    if partitioned and clean_configs:
        for f in os.listdir('/bootpart/boot-versions'):
            _assert(len(f) > 0 and '.' not in f)
            if f in keep: continue
            report(0, 'Deleting old boot version %s' % f)
            rmrf('/bootpart/boot-versions/%s' % f, ignore_errors=True)
    elif clean_configs:
        # 4. Clean up old configs.
        for f in os.listdir('/versions/configs'):
            _assert(len(f) > 0 and not f.startswith('.')) # defensive programming.
            if f == os.path.basename(os.readlink('/versions/boot').rstrip('/')):
                continue
            report(0, 'Deleting old boot configuration %s' % f)
            rmrf('/versions/configs/%s' % f, ignore_errors=True)
    # 5. Clean up updates directory.
    for f in os.listdir('/versions/updates'):
        _assert(len(f) > 0 and not f.startswith('.')) # defensive programming.
        if new_hash and f.startswith(shorten_hash(new_hash)): continue
        report(0, 'Deleting old incomplete update %s' % f)
        rmrf('/versions/updates/%s' % f, ignore_errors=True)

def find_pristine(path, outfile):
    """Write a binary-encoded list of "pristine" inodes in `path` to
    `outfile`."""
    inodes = sorted([int(line.strip()) for line in Popen
                     (['/usr/bin/find',path,'-type','f','-printf','%i\\n'],
                      stdout=PIPE).stdout.readlines()])
    outfile.write(struct.pack('=%dL' % len(inodes), *inodes))

def update_validates(new_hash):
    """Check that hash of contents file is correct and that contents
    file validates."""
    updir = '/versions/updates/%s' % shorten_hash(new_hash)
    cmd = ['/usr/sbin/olpc-contents-verify','%s/etc/passwd' % updir,
           '%s/etc/group' % updir, new_hash, '%s.contents' % updir,
           updir]
    if os.path.exists(updir+'.pristine'):
        cmd.append(updir+'.pristine') # pristine inodes file
    check_call(cmd)
    return True # verification successful!

def check_signature(data, signature, valid_keys, sn):
    """Check a sig01-format signature on the data.  Returns True iff the
    signature is good.
    
    See http://wiki.laptop.org/go/Firmware_Key_and_Signature_Formats#Signature
    for signature specification."""
    try:
        exp = verify_sig(data, signature, valid_keys, sn)
        check_expiration_func()(exp)
    except:
        return False # Bad signature
    # signature is good!
    return True

def check_signed_build(new_root):
    """Check that either the user has a developer key, or the
    downloaded build has a signed OS."""

    # do this import late so that we don't try and read OFW from global import
    # scope
    from bitfrost.leases.keys import OS_KEYS, DEVELOPER_KEYS

    dt = find_devicetree()

    if not os.path.exists(dt + '/mfg-data/wp'): return True # security disabled
    if os.path.exists(dt + '/mfg-data/dk'): return True # perma-dev key
    # read serial number & UUID, so we can check the signature / dev key.
    sn = open(dt + '/mfg-data/SN').read().rstrip("\n\0")
    uuid = open(dt + '/mfg-data/U#').read().rstrip("\n\0")
    if os.path.exists('/security/develop.sig'):
        for devkey in open('/security/develop.sig'):
            try:
                if verify_dev(sn, uuid, devkey, DEVELOPER_KEYS) == 'OK':
                    return True # has a dev key
            except: pass # not a valid dev key.
    # now check signatures on os and ramdisk
    for f in [ os.path.join(new_root,'boot','os.zip'),
               os.path.join(new_root,'boot','rd.zip') ]:
        if not os.path.exists(f): continue # no ramdisk, no worries.
        d = mkdtemp()
        try:
            check_call(['/usr/bin/unzip','-qq',f,'-d',d])
            good_key = False
            certified_data = open(os.path.join(d, 'data.img')).read()
            for signature in open(os.path.join(d, 'data.sig')):
                if check_signature(certified_data, signature, OS_KEYS, sn):
                    good_key = True
            if not good_key: return False
        finally:
            rmtree(d, ignore_errors=True)
    return True # seems to be signed properly

def reboot(poweroff=False):
    """Shutdown and reboot."""
    """
    # CJB says something like this ought to work, but there's probably
    # sugar magic i need to perform.
    from hardware import hardwaremanager
    hw_manager = hardwaremanager.get_manager()
    hw_manager.reboot() # was .shutdown()
    # maybe I should wait a while and call 'shutdown' explicitly?
    """ # XXX see trac #4878
    if poweroff:
        check_call(['/sbin/shutdown','-P','now'])
    else:
        check_call(['/sbin/shutdown','-r','now'])

@with_lock_file('/var/run/olpc-update.pid')
@inhibit_suspend
def perform_update(new_hash, priority, hints, verbosity=0,
                   force=False, poweroff=False):
    """Do initial setup for updates, validate result trees, and install
    successful updates."""
    # see http://wiki.laptop.org/go/Early_Boot#Upgrade_procedure
    def report(lvl, msg):
        if verbosity>=lvl: print >>sys.stderr, msg
    # -1: Check that /versions/pristine/<hash> doesn't already exist.
    if re.match(r'[0-9a-f]+$', new_hash) is None:
        report(0, 'Bad hash: %s' % new_hash)
        sys.exit(1)
    cur_hash = os.path.basename(os.readlink('/versions/running'))
    if shorten_hash(cur_hash) == shorten_hash(new_hash):
        report(0, 'Already running version %s.' % shorten_hash(new_hash))
        sys.exit(1)

    pridir = '/versions/pristine/%s' % cur_hash

    if not os.path.exists(pridir):
        report(0, 'Current version %s doesn\'t exist, perhaps an update is pending and you need to reboot.' % shorten_hash(cur_hash))
        sys.exit(1)

    if os.path.exists('/versions/pristine/%s' % shorten_hash(new_hash)):
        updir = '/versions/updates/%s' % shorten_hash(new_hash)
        upcon = updir+'.contents'
        uppri = updir+'.pristine'
        if os.path.exists('/versions/contents/%s' % shorten_hash(new_hash))\
               and not os.path.exists(updir):
            # Note that this causes a little bit of trouble if the previous
            # image has inappropriate hardlinks, because rsync will write through
            # the hardlink if the only change to the file is permissions.
            # Just Don't Do That.  Use olpc-update 2.7+ and be happy.
            report(0, 'Given version %s already exists; syncing for update.'
                   % shorten_hash(new_hash))
            os.rename('/versions/pristine/%s' % shorten_hash(new_hash),
                      updir)
            os.rename('/versions/contents/%s' % shorten_hash(new_hash),
                      upcon)
            with open(uppri, 'wb') as f:
                find_pristine(updir, f)
    partitioned = _is_partitioned()
    # Create a safety boot configuration
    # steps 0-2: remove old version from boot config.
    trim_config(partitioned, cur_hash)
    # steps 3-5: clean up old or incomplete versions to make space.
    clean_versions(report, partitioned, cur_hash, new_hash)
    # 5b. If /versions/updates/<new hash> doesn't exist, create it.
    updir = '/versions/updates/%s' % shorten_hash(new_hash)
    upcon = updir+'.contents'
    uppri = updir+'.pristine'
    uptmp = mkdtemp(prefix='tmp.', dir='/versions/updates')

    pricon = '/versions/contents/%s' % cur_hash
    if not os.path.exists(uppri):
        with open(uppri, 'wb') as f:
            find_pristine(pridir, f)
    if not os.path.exists(updir):
        os.mkdir(updir)
    # 6-8: OLPC updater.
    #   loop through hints:
    #    make $updir into new updated version,
    #    leave a contents file matching new_hash in
    #       /versions/updates/<new hash>.contents
    import bitfrost.update # parent package has updaters list.
    updaters = vars(bitfrost.update) # all updaters defined in __init__
    for method, url in hints:
        report(0, "Trying %s update from %s" % (method, url))
        method = 'olpc_update_' + method # restrict namespace
        if method not in updaters: continue # unknown update method
        # invoke the update method.  should leave a contents file in place.
        try:
            # updaters should be tolerant of extra keyword arguments.
            updaters[method](vhash=new_hash,
                             old_root=pridir,
                             old_contents=pricon,
                             new_root=updir,
                             new_contents=upcon,
                             tmpdir=uptmp, url=url,
                             verbosity=verbosity)
            # check that hash of contents file is correct and that contents
            # file validates.
            report(0, 'Verifying update.')
            if update_validates(new_hash):
                break # we're done!
        except:
            report(1, "Attempt %s failed: %s" % (method, str(sys.exc_info()[1])))
            # go on to next method.
    else:
        # update failed.
        report(0, "No more update methods to try.")
        sys.exit(2)
    # successful update!
    report(0, 'Installing update in /versions/{pristine,run}/%s'
           % shorten_hash(new_hash))
    # 10. Move /versions/updates/<hash> to /versions/pristine/<hash>
    os.rename(updir, '/versions/pristine/%s' % shorten_hash(new_hash))
    os.rename(upcon, '/versions/contents/%s' % shorten_hash(new_hash))
    os.unlink(uppri)
    rmtree(uptmp, ignore_errors=True)
    #  10b. Create /versions/run/<hash> from /versions/pristine/<hash> according to P_SF_RUN setting
    runtmp = mkdtemp(suffix=shorten_hash(new_hash), dir='/versions/run')
    try:
        check_call(['/bin/cp','-al',
                    '/versions/pristine/%s' % shorten_hash(new_hash),
                    '%s/root' % runtmp])
        # atomically install the 'run' dir.
        os.rename('%s/root' % runtmp,
                  '/versions/run/%s' % shorten_hash(new_hash))
    finally:
        rmtree(runtmp, ignore_errors=True)
    # XXX: someday reimplement P_SF_RUN using COW/ro-bind-mounts, etc.
    # safety check: check that signatures on new image are good. (trac #5307)
    if not force:
        if not check_signed_build('/versions/pristine/%s'
                                  % shorten_hash(new_hash)):
            report(-1, "New build not signed; this update won't boot without "+
                   "a developer key.\nUpdate aborted; use --force to override.")
            sys.exit(3)
    # 11-15: make new boot configuration.
    swing_config(partitioned,
                 current=shorten_hash(new_hash), alt=cur_hash)
    call(['/bin/sync']) # superstition

    # clear the running OS from /versions/pristine now, so that we don't
    # spend so much time doing it during next boot
    report(0, 'Cleaning up...')
    clean_versions(report, partitioned, new_hash, None,
                   clean_run=False, clean_configs=False)

    report(0, 'Update succeeded!')
    #  16. If <priority> reboot.
    if priority=='urgent':
        report(0, "Rebooting now.")
        reboot(poweroff)
    else:
        report(0, 'Restart your system to apply the update.')
    # Done!
    sys.exit(0) # success!
