#!/usr/bin/python
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""A urlopen implementation that supports caching, seek/tell and partial
reads.

This implementation is particularly suitable for use with `zipfile.ZipFile`,
to allow you to read individual files from a remote zip file without
downloading the whole thing."""
from __future__ import division

import httplib
import os
import re
import shutil
import socket
import tempfile
import urllib2
import urlparse
import zipfile

from bitfrost import _assert

_cache = {}
"""Cache of temporary files used to download URLs."""
_redirect_cache = {}
"""Keep track of 'alternative names' for cached URLs."""

MAX_REDIRECTS=5
"""The number of redirects we will follow before assuming we're in a loop."""

def _file_clone(f, length, url, mode='r'):
    """Clone the given file object to a new file object with a new file
    descriptor."""
    # disappointingly os.dup/os.dup2 returns a new file descriptor, but
    # one that shares the file pointer (offset) with the original.
    # so instead we must reopen based on the filename, which is
    # unfortunate.  UNIX needs an openfd call, like QNX has...
    cf = open(f.name, mode)
    # kludge in length() and url() methods.
    return _HttpRangeFile(wrapped_f=cf, wrapped_length=length, wrapped_url=url)

def _tmpprefix(url):
    """Return a maximum of 8 "safe" characters which can be used as a prefix
    for a temporary file name."""
    return re.sub(r'[^a-zA-Z_]', '', url.rsplit('/',1)[-1])[:8]

def _redirect_map(url, redirects=0):
    """Use the redirect_cache to canonicalize the given url."""
    global _redirect_cache, MAX_REDIRECTS
    if url in _redirect_cache or redirects >= MAX_REDIRECTS:
        newurl = _redirect_map(_redirect_cache[url], redirects+1)
        _redirect_cache[url] = newurl # speed up further lookups
        return newurl
    return url

class CancelException(IOError):
    """This exception is raised if a download is cancelled."""
    pass

# Subclass HttpResponse to allow aborting reads.
class HTTPResponse(httplib.HTTPResponse):
    def __init__(self, sock, debuglevel=0, strict=0, method=None):
        httplib.HTTPResponse.__init__(self, sock, debuglevel, strict, method)
        self.cancelhook = lambda: False
    def _safe_read(self, amt):
        s = []
        while amt > 0:
            # THIS IS WHAT WE ADD
            if self.cancelhook():
                raise CancelException()
            # END ADDITION
            chunk = self.fp.read(min(amt, httplib.MAXAMOUNT))
            if not chunk:
                raise httplib.IncompleteRead(s)
            s.append(chunk)
            amt -= len(chunk)
        return ''.join(s)

# Subclass HttpConnection to allow setting the socket timeout.
class HTTPConnection(httplib.HTTPConnection):
    response_class = HTTPResponse
    def connect(self, timeout=None):
        """Set the timeout on the socket before we connect.  Mostly
        cut-and-paste from httplib.py"""
        msg = "getaddrinfo returns an empty list"
        for res in socket.getaddrinfo(self.host, self.port, 0,
                                      socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                self.sock = socket.socket(af, socktype, proto)
                self.sock.settimeout(timeout)
                if self.debuglevel > 0:
                    print "connect: (%s, %s)" % (self.host, self.port)
                self.sock.connect(sa)
            except socket.error, msg:
                if self.debuglevel > 0:
                    print 'connect fail:', (self.host, self.port)
                if self.sock:
                    self.sock.close()
                self.sock = None
                continue
            break
        if not self.sock:
            raise socket.error, msg
    def request(self, method, url, body=None, headers={}):
        # patch upstream to recv() after we send, so that a closed
        # connection is discovered here rather than later in
        # HTTPResponse._read_status()
        try:
            self._send_request(method, url, body, headers)
            b = self.sock.recv(1, socket.MSG_PEEK)
            if b == '':
                self.close()
                raise socket.error(32) # sender closed connection.
        except socket.error, v:
            # trap 'Broken pipe' if we're allowed to automatically reconnect
            if v[0] != 32 or not self.auto_open:
                raise
            # try one more time
            self._send_request(method, url, body, headers)


# subclass HTTPHandler to use our HTTPConnection
def HTTPHandler(timeout=None, cancelhook=(lambda: False)):
    """Returns a subclass of `urllib2.HTTPHandler` which sets the
    socket timeout to the given `timeout`.  Works by creating a
    subclass of `HTTPConnection` with the desired timeout, and then
    creating a subclass of `urllib2.HTTPHandler` which uses that
    `HTTPConnection`."""
    class HC(HTTPConnection):
        def connect(self):
            HTTPConnection.connect(self, timeout)
        def getresponse(self):
            res = HTTPConnection.getresponse(self)
            res.cancelhook = cancelhook
            return res
    class _HTTPHandler(urllib2.HTTPHandler):
        handler_order = 499 # before the default HTTP handler
        def http_open(self, req):
            return self.do_open(HC, req)
    return _HTTPHandler

class _HttpRangeFile(object):
    """Return a file-like object which supports seek/tell via byte-range
    requests.  Will bail to an implementation which downloads the entire
    url to a temporary file and does seek/tell on it, if the target
    http server doesn't support partial GETs."""
    def __init__(self, url=None, wrapped_f=None, wrapped_length=None,
                 wrapped_url=None, timeout=None):
        if url is not None:
            self.f = _InnerWrapper(self, url, timeout)
            self._already_set_up = False # no connection yet.
        else:
            self._already_set_up = True
            self.f = wrapped_f
            self._length = wrapped_length
            self._url = wrapped_url
    def seek(self, offset, whence=0):
        self.f.seek(offset, whence)
    def tell(self):
        return self.f.tell()
    def read(self, size=-1):
        return self.f.read(size)
    def _get_url(self):
        """Returns the (possibly redirected) url corresponding to this
        file-like object."""
        self._setup() # make sure _url is initialized
        return self._url
    # use 'url' property like urllib2.urlopen gives.
    url = property(_get_url, doc=_get_url.__doc__)
    def length(self):
        """Can throw `urllib2.HTTPError` if url is not found."""
        self._setup() # make sure _length is initialized
        return self._length
    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None
    def _setup(self):
        if not self._already_set_up: self.f._setup()
    # context manager support.
    def __enter__(self):
        return self
    def __exit__(self, *exc_info):
        """Closes the file."""
        self.close()
    def __del__(self):
        self.close()

class _InnerWrapper(object):
    """This file-like object uses byte-range requests of the remote HTTP
    server.  It bails to a full download, redirecting its outer wrapper,
    if byte-range requests fail, or if the entire file is requested."""
    def __init__(self, outer, url, timeout=None, redirects=0):
        self.outer, self.url, self.conn, self.offset = outer, url, None, 0
        self.accepts_ranges = True
        self.timeout = timeout
        self.redirects = redirects
    def _setup(self):
        if self.conn is not None: return # already set up
        pieces = urlparse.urlsplit(self.url)
        if pieces.scheme != 'http':
            # bail: GET to a temp file, and then open the temp file.
            return self._bail()
        self.conn = HTTPConnection(pieces.netloc)
        # do a Content-Length transaction; bail if the server doesn't play nice
        self.path = urlparse.urlunsplit \
                    ( (None, None, pieces[2], pieces[3], pieces[4]) )
        # be nice to broken URLs which start with /./ or /../ or //
        self.path = re.sub(r'^/(\./|\.\./|/)+', '/', self.path)
        res = None
        try:
            self.conn.connect(timeout=self.timeout)
            req = self.conn.request('HEAD', self.path,
                                    headers={'Connection': 'Keep-Alive'})
            res = self.conn.getresponse()
        except:
            return self._bail()
        # handle redirect request.  implementation borrowed from urllib2.py
        if self.redirects < MAX_REDIRECTS and \
               res.status in (301, 302, 303, 307):
            global _redirect_cache
            # Some servers (incorrectly) return multiple Location headers
            # (so probably same goes for URI).  Use first header.
            l = [v for h, v in res.getheaders() if h in ('location', 'uri') ]
            if not l: return self._bail()
            newurl = urlparse.urljoin(self.url, l[0])
            # be nice to broken clients which add spaces to the URIs
            newurl = newurl.replace(' ', '%20')
            self.conn.close()
            self.conn = None
            _redirect_cache[self.url] = newurl
            self.url = newurl
            self.redirects += 1
            return self._setup()

        if res.status != httplib.OK:
            return self._bail()
        length = res.getheader('content-length', None)
        if length is None:
            return self._bail()
        self.length = self.outer._length = int(length)
        # verify that server accepts ranges
        if res.getheader('accept-ranges', None) != 'bytes':
            # boo.  don't bail yet, but bail on first read.
            self.accepts_ranges = False # boo
        # look at etag
        etag = res.getheader('etag', None)
        if etag is not None and etag.startswith('W/'):
            etag = None # weak etags are useless
        self.etag = etag if etag is not None else '*'
        # read (should be zero-length)
        res.read()
        # okay, we're good to go!
        self.outer._url = self.url # possibly redirected by now.
        self.outer._already_set_up = True
        return
    def seek(self, offset, whence=0):
        if whence == 0:
            self.offset = offset
        elif whence == 1:
            self.offset += offset
        elif whence == 2: # offset negative in this case
            self._setup() # ensure self.length is initialized
            self.offset = self.length + offset
    def tell(self):
        return self.offset
    def read(self, size=-1):
        if size == 0: return [] # optimization
        if size < 0:
            if self.offset == 0: # optimization
                self._bail() # read whole file & cache it for the future
                return self.outer.read()
            self._setup()
            size = self.length - self.offset
        self._setup()
        _assert(self.offset <= self.length)
        if (self.offset + size) > self.length:
            size = self.length - self.offset # fixup overlong requests
        if size == 0: return [] # we're at the end of our rope.
        if not self.accepts_ranges:
            # we actually need to do a partial read, we have to bail.
            self._bail()
            return self.outer.read(size)
        # okay, do a byte-range request.
        req = self.conn.request\
              ('GET', self.path,
               headers={'Range': 'bytes=%d-%d' % \
                        (self.offset, self.offset+size-1),
                        'If-Match': self.etag,
                        'Connection': 'Keep-Alive'})
        res = self.conn.getresponse()
        if res.status == httplib.OK:
            # hrm, server is ignoring the 'Range' request, bail!
            self._bail()
            return self.outer.read(size)
        elif res.status != httplib.PARTIAL_CONTENT:
            # some other problem, throw it.
            self.close()
            raise IOError('could not read: %s' % res.reason)
        data = res.read(size)
        _assert(len(data) == size)
        self.offset += size
        res.close()
        return data
    def close(self):
        if self.conn is not None:
            self.conn.close()
            self.conn = None
        self.outer = None # unlink from outer wrapper
    def __del__(self):
        self.close()
    def _bail(self):
        global _cache, _redirect_cache
        # our http server doesn't support byte serving; download the full file
        # and serve from the file instead.
        if self.conn is not None:
            self.conn.close()
            self.conn = None
        tf = tempfile.NamedTemporaryFile(prefix=_tmpprefix(self.url))
        try:
            opener = urllib2.build_opener(HTTPHandler(self.timeout))
            uo = opener.open(self.url) # throws URLError on any problem
            if uo.url != self.url:
                _redirect_cache[self.url] = uo.url
                self.url = uo.url
            shutil.copyfileobj(uo, tf)
            uo.close()
        except urllib2.URLError:
            _cache[self.url] = None # cache the failure.
            raise
        tf.seek(0, 2) # go to the end
        self.outer._length = tf.tell()
        # record this temp file in our cache.
        _cache[self.url] = tf
        # reopen in read mode, reset to previous location
        tf = _file_clone(tf, tf.tell(), self.url) # reopen in read mode.
        tf.seek(self.offset) # reset to match where we 'were'
        self.outer.f = tf # unlink this Wrapper from its container
        self.outer._url = self.url
        self.outer._already_set_up = True

def urlopen(url, use_cache=True, timeout=None):
    """Return a file-like object which accesses the given URL in a way
    which supports seek/tell and partial reads.  If `use_cache` is
    True (which is the default) will store any URLs which end up being
    read completely to a persistent cache to satisfy future requests."""
    global _cache
    if use_cache:
        url = _redirect_map(url)
    if use_cache and url in _cache:
        if _cache[url] is None:
            raise urllib2.URLError("Replaying exception")
        return _file_clone(_cache[url], _cache[url].tell(), url)
    return _HttpRangeFile(url, timeout=timeout)

def urlretrieve(url, filename=None, dir=None,
                reporthook=(lambda size, ttlsize: None),
                cancelhook=(lambda: False),
                use_cache=True, store_cache=True,
                timeout=None):
    """Retrieve a url to a temporary file and return the name of the
    temporary file.  Arguments are similar to `urllib.urlretrieve` but
    the return value is just the filename of the fetched file (not a
    tuple as in `urllib.urlretrieve`).  The `reporthook` gets two
    arguments: size and total size. The first invocation has size==0
    and the last one has size==total size.  The total size parameter will
    be None for all but the final invocation if the remote host does not
    send a proper 'Content-Length' header.  The `cancelhook` should
    return True if the download is to be aborted, in which case
    `CancelException` will be raised.  The `use_cache` parameter
    determines whether `urlretrieve` can/should use data already
    cached from previous calls to `urlopen` and `urlretrieve`, and the
    `store_cache` parameter determines whether downloaded files which
    were not already in the cache should be added to the cache.
    The `dir` parameter names a directory to use for the temporary file,
    if present (and if `filename` is not present)."""
    global _cache, _redirect_cache
    BLOCK_SIZE = 64*1024
    if use_cache:
        url = _redirect_map(url)
    if filename is None:
        fdst, outname = tempfile.mkstemp(dir=dir, prefix=_tmpprefix(url))
        fdst = os.fdopen(fdst, 'wb', BLOCK_SIZE)
    else:
        fdst, outname = open(filename, 'wb', BLOCK_SIZE), filename
    try: # after this point, we should clean up by deleting outname
        if use_cache and url in _cache:
            fsrc = urlopen(url, use_cache=use_cache, timeout=timeout)
            length = fsrc.length()
        else: # for efficiency, don't do partial reads.
            opener = urllib2.build_opener(HTTPHandler(timeout, cancelhook))
            fsrc = opener.open(url)
            if store_cache and fsrc.url != url:
                _redirect_cache[url] = fsrc.url
                url = _redirect_map(url)
            length = fsrc.headers.getheader('Content-Length')
            if length is not None: length = long(length)
        fcache = None if (url in _cache or not store_cache) else \
                tempfile.NamedTemporaryFile(dir=dir, prefix=_tmpprefix(url))
        read = 0
        while True:
            reporthook(read, length)
            if cancelhook(): # abort!
                raise CancelException()
            buf = fsrc.read(BLOCK_SIZE)
            if not buf: break
            fdst.write(buf)
            if fcache is not None: fcache.write(buf)
            read += len(buf)
        reporthook(read, read)
        if fcache is not None:
            _cache[_redirect_map(url)] = fcache
        return outname
    except:
        os.unlink(outname) # clean up
        raise
        

def urlcleanup():
    """Clear the cache which may have been built up by previous calls to
    `urlopen` and `urlretrieve`."""
    global _cache, _redirect_cache
    _cache = {} # temp files are closed & unlinked as they become dereferenced
    _redirect_cache = {}

def _main():
    """self-test."""

    # this sequence of partial reads fails when going through a squid proxy.
    url = 'http://dev.laptop.org/~sayamindu/bundles/read/Read-49.xo'
    uo = urlopen(url)
    uo.seek(120863)
    uo.read(22)
    uo.seek(104512)
    uo.read(16351)
    uo.seek(102823)
    uo.read(30)

    # test that we can read the directory of a bundle w/o downloading all of it
    url = 'http://dev.laptop.org/~cscott/Firefox-1.xo?test#foo'
    if True:
        zf = zipfile.ZipFile(urlopen(url), 'r')
        zf.printdir()
        zf.close()
        print "----"

    # test redirects
    url = 'http://etoys.laptop.org/xo'
    if True:
        _assert(urlopen(url).url == (url+'/'))
        # check that cached copy is stored with the canonical url
        saved = urlretrieve(url)
        print saved
        os.unlink(saved)
        _assert(url in _redirect_cache)
        _assert(_redirect_cache[url] in _cache)

    # test the progress callbacks.
    import sys
    def progress(cursize, totalsize):
        if totalsize is None:
            sys.stdout.write('%d bytes\r' % cursize)
        else:
            sys.stdout.write('%d bytes of %d total (%3.1f%%)\r' % (cursize, totalsize, 100*(cursize / totalsize)))
        sys.stdout.flush()
    # test fetching a file with no Content-Length given by the server.
    if True:
        saved = urlretrieve('http://cscott.net', reporthook=progress)
        print
        print saved
        os.unlink(saved)
    # test downloading a large file
    url = 'http://dev.laptop.org/~cscott/Firefox-1.xo?test#foo'
    if True:
        saved_xo = urlretrieve(url, reporthook=progress, use_cache=True)
        print
        print saved_xo
    # test that second download comes from cache
    if True:
        saved_xo = urlretrieve(url, reporthook=progress)
        print
        print saved_xo
    # and that we can create an ActivityBundle from the download
    if True:
        from sugar3.bundle.activitybundle import ActivityBundle
        ab = ActivityBundle(saved_xo)
        print ab.get_name()
    # clean up
    os.unlink(saved_xo)

if __name__=='__main__': _main()
